/*
 * maplib - library routines to draw a map from a .MAP file.  These files
 * are generated by the "map" program using the worldmap database.
 *
 * The user must call
 *	graphics();
 *	erasegraph();
 * before using this package, and
 * 	text();
 * afterwards, to switch the display to and from graphics mode.
 *
 * Public entry points are as follows:
 * plotmap(fname, grid) - plot the file named fname with lines every grid
 *     degrees.  If grid is zero, don't draw a grid.
 *
 * mapline(lat, lon) - draw a line from current position to (lat, lon)
 *
 * mapmove(lat, lon) - move current position to (lat, lon) w/o drawing
 *
 * maplabel(fmt, <args>) - printf a label at the current position
 *
 * By E. Roskos 3/89
 *
 */ 
 
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <math.h>

/*
 * max display coordinate allowed for either an X or a Y coordinate.  Assumes
 * a 1:1 aspect ratio; my HGC library scales coordinates to give a 1:1
 * ratio, so the maximum y value is greater than the number of raster lines.
 */
#define XYMAX	635

/*
 * display point around which Y coordinates are inverted (because HGC 
 * coordinates are upside-down from cartesian coordinates)
 */
/* #define YMAX	550 */

#define YMAX 	ymax

/*
 * scaling margin: this total number of pixels is guaranteed to be
 * present around the border of the map, for labels, etc. to go in.
 */
#define MARGIN	20

/*
 * amount of margin to add on left of the display.  This is just to
 * center the picture in the screen, since the physical screen area
 * (adjusted for the square aspect ratio) is rectangular.
 */
#define XOFFSET 15

/*
 * the basic coordinate data structure of the program.  A single coordinate
 * is stored in lat and lon; fnew is set to 1 if the pen is supposed to be
 * up when moving to this coordinate, or to 0 if it is supposed to be down
 * when moving to this coordinate.  fnew will also be set to -1 when
 * this structure is returned from a file read, if EOF is detected.
 *
 * Binary files are simply files of coord.
 *
 * The latitude (lat) and longitude (lon) are always represented in
 * units of degrees (the normal world-globe degrees).
 */
typedef struct coord
{
	float lat;	/* latitude */
	float lon;	/* longitude */
	int fnew;	/* 0=draw line, 1=just move pen, -1=EOF */
} coord;

/*
 * ymax is the coordinate of the bottom line of the map.  It is adjusted
 * to center the map vertically on the display.
 */
static int ymax;

/*
 * offmap - set TRUE by draw() if last coord was off the map.
 * This is an important global; it's used to suppress spurious lines
 * when the cursor's current position is invalid because the previous
 * point was off the map, meaning a mapline() while in this state would
 * draw the line starting from the wrong point.
 */
static int offmap = 0;

/* 
 * flag is set true to tell getcoord we're starting a new file (and
 * thus that the pen should be up for the first coord read).
 */
static int fnewfile = 1;

/*
 * getcoord(f) - reads the next coord structure from the input file.
 */
static coord
getcoord(f)
FILE *f;
{
coord cd;
char *cp;

	fread(&cd, sizeof(cd), 1, f);
	if (feof(f)) cd.fnew = -1;
	if (fnewfile)
		cd.fnew = 1;
	fnewfile = 0;
	return(cd);
}

/*
 * isbinary(s)
 * returns true if s is the name of a binary file, i.e., if it
 * ends in the extension .map - otherwise, returns FALSE.
 */
static int
isbinary(s)
char *s;
{
char *p;

	p = strchr(s, '.');
	if (p==NULL) return(0);
	p++;
	if (strcmp(strlwr(p),"map")==0) return(1);
	else return(0);
}

/*
 * scaling parameters
 * These contain the range of latitudes and longitudes seen so far
 * (across arbitrarily many input files), and (after setscale is
 * called) also contain the difference between the max and min of
 * each, and the scaling factor to use.  In considering the scaling,
 * remember that my HGC library automatically scales its coordinates
 * to give a 1:1 physical aspect ratio on the screen. 
 */
static float latmin = 9999, latmax = -9999;	/* min and max latitude  */
static float lonmin = 9999, lonmax = -9999;	/* min and max longitude */
static float latd, lond;	/* diff. between max and min latitude and longitude */
static float scaling;		/* number of HGC pixels per degree */

/*
 * assess(f)
 * Read the header information stored in binary map file f
 * to determine the minimum and maximum latitude and longitude 
 * in the file.
 */
static void
assess(f)
FILE *f;
{
coord cd;

	fread(&latmin, sizeof(latmin), 1, f);
	fread(&latmax, sizeof(latmax), 1, f);
	fread(&lonmin, sizeof(lonmin), 1, f);
	fread(&lonmax, sizeof(lonmax), 1, f);
}

/*
 * setscale()
 * Based on the globals set by assess(), compute the proper scaling
 * factor to make the assessed set of map files fill the display to
 * the extent possible.
 */
static void
setscale()
{
int yd;

	latd = latmax - latmin;
	lond = lonmax - lonmin;
	
	/* 
	 * compute pixels per degree.  This depends on whether the
	 * latitude or longitude delta is larger, and is computed so
	 * the map aspect ratio stays square on the display.
	 * The constant MARGIN that is subtracted from XYMAX is to give
	 * a guaranteed margin around the map for meridian labelling.
	 */
	scaling = (float)(XYMAX-MARGIN) / (latd > lond? latd : lond);

	yd = (int)(latd * scaling);
	
	/*
	 * center the map on the display
	 */
	ymax = XYMAX - ((XYMAX - yd) / 2);
}

/*
 * draw(cd)
 * Execute a coord as a graphics command.  
 * See the definition of type coord.
 */ 

static void
draw(cd)
coord cd;
{
int x, y;
float fx, fy;

#ifdef DEBUG
	printf("(lat,long) = (%10.5f, %10.5f) ",cd.lat, cd.lon);
#endif
	/*
	 * ignore the coord if it's off the map
	 */
	if (cd.lat > latmax || cd.lat < latmin ||
	    cd.lon > lonmax || cd.lon < lonmin)
	{
#ifdef DEBUG
		printf("off map [lat (%f, %f) lon (%f, %f), ignored\n",
			latmin, latmax, lonmin, lonmax);
#endif
		fnewfile = 1;	/* don't draw anything */
		offmap = 1;
		return;
	}
	
	offmap = 0;
	
	fy = (cd.lat - latmin) * scaling;
	fx = (cd.lon - lonmin) * scaling;
#ifdef DEBUG
	printf("scaled point = (%10.5f, %10.5f); ", fx, fy);
#endif
	x = (int)(fx+0.5);
	y = (int)(fy+0.5);
#ifdef DEBUG
	printf("HGC coord = (%d, %d)\n", x, y);
#endif
	if (cd.fnew) moveto(x + XOFFSET, YMAX - (y+(MARGIN>>1)));
	else lineto(x + XOFFSET, YMAX - (y+(MARGIN>>1)));
}

/*
 * grid(inc)
 * draw a latitude/longitude grid line every inc degrees, aligned
 * on absolute multiples of inc.  This is done by constructing
 * coords to draw the grid and executing them with draw.
 */
static void
grid(inc)
int inc;
{
float fx, fy;	/* current lat or lon line being drawn */
float mx, my;	/* saved 1st lat and lon values */
int x, y;	/* actual screen coordinates, used to draw labels */
coord cd;	/* coord, used to generate the lines */

	/* set hershey font parameters */
	textscale(1, 1);
	textfont(0);
	textleading(25);

	/* compute first latitude */
	fx = (float)(((int)latmin / inc) * inc);
	if (fx < latmin) fx += inc;
	mx = latmin;

	/* compute first longitude */
	fy = (float)(((int)lonmin / inc) * inc);
	if (fy < lonmin) fy += inc;
	my = lonmin;
	
	/* plot the latitude lines */
	while (fx < latmax)
	{
#ifdef MSDOS
		/* make grid grey */
		dotline(1);
#endif
		cd.lat = fx;
		cd.lon = my;
		cd.fnew = 1;
		draw(cd);
		cd.lon = lonmax;
		cd.fnew = 0;
		draw(cd);

		/* label the line */
#ifdef MSDOS
		dotline(0);
#endif
		x = getx();
		y = gety();
		moveto(x+10, y);
		hprintf("%.0f%c", fabs(fx), fx<0? 'S' : 'N');

		/* go to next line */
		fx += (float)inc;
	}	

	/* plot the longitude lines */
	while (fy < lonmax)
	{
#ifdef MSDOS
		/* make grid grey */
		dotline(1);
#endif
		cd.lat = mx;
		cd.lon = fy;
		cd.fnew = 1;
		draw(cd);
		cd.lat = latmax;
		cd.fnew = 0;
		draw(cd);

		/* label the line */
#ifdef MSDOS
		dotline(0);
#endif
		x = getx();
		y = gety();
		moveto(x-15, y-15);
		hprintf("%.0f%c", fabs(fy), fy<0? 'W' : 'E');

		/* go to next line */
		fy += (float)inc;
	}	
}

/*
 * PUBLIC
 * plotmap(fname, grid)
 * Read in file named fname and plot it on the display.
 * If gridsz is nonzero, draw a grid with the specified spacing.
 * Returns 1 if plotted successfully, 0 if error.
 */ 
int
plotmap(fname, gridsz)
char *fname;
int gridsz;
{
coord cd;
FILE *f;

	if (!isbinary(fname))
	{
		fprintf(stderr, "\007File `%s' does not end in .MAP\n", fname);
		return(0);
	}
	
	f = fopen(fname, "rb");
	if (f==NULL)
	{
		fprintf(stderr, "\007");
		perror(fname);
		return(0);
	}

	assess(f);
	setscale();
	if (gridsz) 
	{
		color(5);
		grid(gridsz);
	}

	fnewfile++;	/* tell getcoord to start with pen-up */
	fseek(f, 4L*sizeof(float), 0); /* rewind to just past header */

	color(3);

	while (cd = getcoord(f), cd.fnew != -1)
	{
		draw(cd);
	}

	fclose(f);
	return(1);
}

/*
 * PUBLIC
 * mapline(lat, lon)
 * Draw a line from the current position to coordinate (lat, lon)
 */
void
mapline(lat, lon)
float lat, lon;
{
coord cd;

	cd.lat = lat;
	cd.lon = lon;
	cd.fnew = offmap? 1 : 0; /* just move if last point was off map */
	
	color(1);
	draw(cd);
}	

/*
 * PUBLIC
 * mapmove(lat, lon)
 * Move the current position to coordinate (lat, lon)
 */
void
mapmove(lat, lon)
float lat, lon;
{
coord cd;

	cd.lat = lat;
	cd.lon = lon;
	cd.fnew = 1;
	
	draw(cd);
}

/*
 * PUBLIC
 * maplabel(fmt, <args>)
 * Label the current point on the map, to the right of the point
 */
void
maplabel(fmt, a, b, c, d, e, f, g, h, i, j, k, l)
char *fmt;
int a, b, c, d, e, f, g, h, i, j, k, l;
{
int savx, savy;

	/* don't label if last point was off the map */
	if (offmap)
		return;
		
	/* set hershey font parameters */
/** jv 13, 16 **/
	textscale(21,16);
	textfont(0);
	textleading(15);
	dotline(0);	

	/* save where we currently are */
	savx = getx();
	savy = gety();
	
	/* draw a square to mark the point we're labelling */
	color(4);
	moveto(savx-2, savy-2);
	lineto(savx+2, savy-2);
	lineto(savx+2, savy+2);
	lineto(savx-2, savy+2);
	lineto(savx-2, savy-2);
	
	/* write the labelling text */
	color(7);
	moveto(savx+8, savy+3);
	hprintf(fmt, a, b, c, d, e, f, g, h, i, j, k, l);
	
	/* move back to where we were originally */
	moveto(savx, savy);
}
	

/*
 * SIGINT (^C or BREAK) trap handler to restore the display
 * to text mode.
 */
static void
terminate()
{
	text();
	exit(2);
}		

#ifdef TEST

/*
 * the main program, for debugging only
 */
main(argc, argv)
int argc;
char **argv;
{

	if (argc != 2)
	{
		printf("usage: maplib mapname\n");
		text();
		exit(1);
	}
	
	graphics();
	erasegraph();
	plotmap(argv[1], 15);
	mapmove(35.0, -89.0);
	maplabel("35 -89\nFIRST LABEL");
	mapline(40.0, -70.0);
	maplabel("40 -70\nSECOND LABEL"); 
	mapline(50.0, -65.0);
	maplabel("65 -65\nTHIRD LABEL");
	printf("\007");
	while (getchar() != '\n') ;
	text();
	exit(0);	
}

#endif
